/********************************************************************

	+--------------------------------------------------------+
	|                    Site Swap Scribe                    |
	|                                                        |
	|           Site swap --> PostScript converter           |
	+--------------------------------------------------------+

	(c) Y. King / Summer 1992 / Vancouver, B.C.
	Permission is granted to make any use of this file provided:
	    1) You don't make any money for it without my permission
	    2) You don't get *me* in trouble by doing something illegal
	       with it
	    3) This notice is included in all copies
	    4) Altered copies are marked as such

	Any bugs, let me know; or even if there aren't any bugs, I'd
	be interested to hear that you're using it ... I'm at
	<yking@cs.ubc.ca> (at least until December '92).

 ********************************************************************/

#include	<stdio.h>
#include	<memory.h>

	/* File and path for PostScript prologue stuff. The path
	   should really be an environment variable, but getenv()
	   varies wildly across compilers. (BTW, NULL path means
	   use current directory.) Make sure the trailing slash
	   is part of the pathname. */
char	*psfile = "sss.ps";
char	*pspath = NULL;		/* NULL means current directory */

#define	DIR_SLASH	'/'

	/* These are "standard" macros, but it's easier to rewrite them
	   than to find out where they are for a whole bunch of compilers.
	   These definitions may not precisely mimic the standard ones,
	   but they do the job. (Note that I've also rewritten stricmp
	   and strnicmp, for similar reasons.) */
#define	ISALPHA(C)	(('a' <= (C) && (C) <= 'z') || ('A' <= (C) && (C) <= 'Z'))
#define	ISDIGIT(C)	(('0' <= (C) && (C) <= '9'))
#define	ISSPACE(C)	((C) == ' ' || (C) == '\t' || (C) == '\n')

	/* Definitions (most units in inches) */
#define	TOP_MARGIN	0.5
#define	BOTTOM_MARGIN	0.25
#define	TITLE_SPACE	0.75
#define	LINE_WIDTH	0.4

#define	BUF_SIZE	2048

	/* PostScript units are 1/72nd of an inch */
#define	PSCOORD(X)	((X)*72.0)

	/* Only print stuff if it's wanted (a little ugly, but it works) */
#define	PSPRINT(X)	if (PSOut) X

	/* Some cheesy little macros */
#define	ABS(X)		((X) < 0 ? (-(X)) : (X))
#define	MAX(X,Y)	((X) < (Y) ? (Y) : (X))
#define	MIN(X,Y)	((X) > (Y) ? (Y) : (X))

	/* Max throws per pattern. This is used for an array bound
	   which is never checked, so there's potential for segmentation
	   faults here. */
#define	MAX_THROWS	500

	/* Hands */
#define	NUM_HANDS	2
#define	R		1
#define	L		0

	/* Structures */
typedef	struct pattern
	{
		int	nBalls;
		int	nThrows;
		int	sumThrows;
		int	nReps;		/* i.e. cycles through pattern */
		int	Throw[MAX_THROWS];	/* The actual pattern */

		struct pattern	*next;	/* for pattern lists */
	}
		PATTERN;

typedef	struct pageloc
	{
		int	hand, row, col, page;
	}
		PAGELOC;

typedef	struct backlog	/* for storing draw_throw commands */
	{
		PAGELOC	src, dest;

		struct backlog	*next;
	}
		BACKLOG;

	/* Prototypes (non-ANSI) */
void	main();

void	read_file();
void	read_comline();
PATTERN	*start_pattern();
void	add_pattern();

void	plcopy();
int	plequal();
int	plinc();

PATTERN	*new_pattern();

void	draw_page();
void	draw_numbers();
void	draw_throw();
void	add_overflow();
void	start_page();
void	cycle_sym();
void	edit_sym();
void	edit_check();

void	line();
void	linewidth();
void	circle();
void	curve();
void	drawball();
void	ballcolour();
void	font();
void	text();
void	outlinetext();
void	box_text();
void	clip();
void	unclip();
void	showpage();

void	warning();

int	mystricmp();
int	mystrnicmp();

char	buf[BUF_SIZE];	/* for string stuff */

	/* Global variables ... bad programmer! naughty programmer! */
char	*title = NULL;
char	*infile = NULL;
char	*outfile = NULL;
int	PSOut = 1;

float	PageWidth = 8.5;
float	PageHeight = 11.0;

float	ColTop;
float	ColBot = BOTTOM_MARGIN;
float	RowHeight = -1.0;
float	ColWidth = -1.0;
int	nCols = 2;
int	nRows = -1;
float	ColSpace;
float	*ColCenter;
float	*ColLeft;
float	*ColRight;
float	Height, srcHeight, destHeight;
float	*ColHandX[NUM_HANDS];
float	BallSize = 0.3;
int	nPages = 1;

float	*ballColour;		/* Ball colours */

int	colour = 0;		/* Use colour? */
int	shading = 0;		/* Or gray shades? */
int	numLabs = 1;		/* Number each throw */
int	ballNums = 1;		/* Put siteswap numbers on balls */
int	debug = 1;		/* Print warning messages */
int	editMarks = 1;		/* Put edit symbols on errors */
int	cycleMarks = 1;		/* Mark cycle changes */
int	autoStart = 0;		/* Generate startup pattern */
int	upsideDown = 0;		/* Start pattern at bottom of page */

BACKLOG	**overflow;	/* Overflows from one page to next */

int	throwNum = 0;
int	nThrowsTotal;	/* Total throws to be printed */
int	*BallPresent;	/* For error checking */

int	nCycles = -1;	/* Total times through entire pattern.
			   Less than zero means nPages is limit. */

int	nBalls = 0;			/* Total number of balls */
int	init[NUM_HANDS] = {-1,-1};	/* Initial balls in each hand */

PATTERN	*Startup = NULL;	/* Startup pattern */
PATTERN	*Pattern = NULL;	/* And a list of patterns */

/*-------------------------------------------------------*/
void	main(argc, argv)

	int	argc;
	char	*argv[];
{
	PATTERN	*P;

	PAGELOC	src, dest;
	int	i,j;

		/* Get options from command line (this will also
		   handle reading from (nested) files). */
	read_comline(argc, argv);

	if (title == NULL)
		title = "Site Swap Scribe";

		/* Initial pattern gets special treatment.The pattern
		   is checked for validity. If the number of balls in
		   each hand is not set in input file then it is deduced
		   from the total number of balls.

		   If no startup is given and auto generation
		   requested then it is done (if possible). */
	if (Pattern != NULL)
	{
		P = start_pattern(Pattern);

		if (autoStart)
		{
			if (P == NULL)
				warning("no autostart possible; ignoring -autostart");
			else if (Startup != NULL || init[L] >= 0 || init[R] >= 0)
				warning("startup given; ignoring -autostart");
			else if (P->nThrows == 0)
			{
				fprintf(stderr, "no startup necessary\n");
			}
			else
			{
				fprintf(stderr, "startup generated: -s ");
				for (i = 0; i < P->nThrows; i++)
					fprintf(stderr, "%-d ", P->Throw[i]);
				fprintf(stderr, "\n");

				Startup = P;
			}
		}

		if (P != NULL)
		{
			init[R] = P->nBalls/2 + P->nBalls%2;
			init[L] = P->nBalls/2;
			nBalls = P->nBalls;
		}
	}

	if (nBalls == 0 && init[R] >= 0 && init[L] >= 0)
		nBalls = init[R] + init[L];

		/* Set up the output file */
	if ((outfile != NULL) && PSOut)
	{
		if (freopen(outfile, "w", stdout) == NULL)
		{
			fprintf(stderr, "couldn't open %s\n", outfile);
			exit(1);
		}
	}

		/* Postscript prologue */
	if (PSOut)
	{
		FILE	*psf;

		if (pspath != NULL)
		{
			strcpy(buf,pspath);
		}
		else
			buf[0] = 0;

		strcat(buf, psfile);

		if ((psf = fopen(buf, "r")) == NULL)
		{
			fprintf(stderr, "couldn't open PostScript prologue %s\n", buf);
			exit(1);
		}

		while (!feof(psf))
			if(fgets(buf, BUF_SIZE, psf) != NULL)
				fputs(buf, stdout);

		fclose(psf);
	}

		/* Initialize screen specs */
	ColTop = PageHeight - TOP_MARGIN - TITLE_SPACE;
	if (ColWidth < 0)
		ColWidth = (PageWidth - nCols*0.5)/(float)nCols;
	ColSpace = (PageWidth-ColWidth*nCols)/(float)(nCols + 1);
	if (RowHeight < 0.0)
	{
		if (nRows < 0)
			nRows = 25;
		RowHeight = (ColTop - ColBot)/(float)nRows;
	}
	else
	{
		if (nRows < 0)
			nRows = (ColTop - ColBot)/RowHeight;
		ColBot = ColTop - nRows*RowHeight;
	}

	BallSize *= RowHeight;

	ColCenter = (float *) malloc(nCols*sizeof(float));
	ColLeft = (float *) malloc(nCols*sizeof(float));
	ColRight = (float *) malloc(nCols*sizeof(float));
	ColHandX[R] = (float *) malloc(nCols*sizeof(float));
	ColHandX[L] = (float *) malloc(nCols*sizeof(float));

	overflow = (BACKLOG **) malloc(nPages*sizeof(BACKLOG *));
	for (i = 0; i < nPages; i++)
		overflow[i] = NULL;

	for (i = 0; i < nCols; i++)
	{
		ColLeft[i] = i * (ColWidth+ColSpace) + ColSpace;
		ColCenter[i] = ColLeft[i] + ColWidth/2.0;
		ColRight[i] = ColLeft[i] + ColWidth;
		
		ColHandX[R][i] = ColCenter[i] + ColWidth*0.25;
		ColHandX[L][i] = ColCenter[i] - ColWidth*0.25;
	}

	draw_page(0);
	draw_numbers(Pattern);

		/* Initialize source position on page
		   (start in second row [i.e. row 1]) */
	src.hand = R;
	src.row = 1;
	src.col = 0;
	src.page = 0;

		/* Set up ongoing error checking stuff */
	if (nCycles < 0)
	{
			/* first row/first col used already */
		nThrowsTotal = nPages*nCols*nRows;
	}
	else
	{
		nThrowsTotal = 0;
		for (P = Pattern; P != NULL; P = P->next)
		{
			nThrowsTotal += P->nThrows*P->nReps;
		}
		nThrowsTotal *= nCycles;
		if (Startup != NULL)
			nThrowsTotal += Startup->nThrows;
	}

	BallPresent = (int *) malloc(nThrowsTotal*sizeof(int));

	for (i = 0; i < nThrowsTotal; i++)
		BallPresent[i] = 0;

		/* If pattern defined then put balls in hands */
	if (Pattern != NULL)
	{
		ballColour = (float *) malloc(MAX(1,nBalls)*sizeof(float));
		for (i = 0; i < nBalls && i < nThrowsTotal; i++)
		{
			ballColour[i] = (float) i/(float) nBalls;
			BallPresent[i] = i + 1;
		}
	}

		/* Output the startup sequence */
	if (Startup != NULL && Startup->nThrows > 0)
	{
		cycle_sym("startpatternsym", &src);

		for (i = 0; i < Startup->nThrows; i++)
		{
			edit_check(&src, Startup->Throw[i]);

			plcopy(&dest,&src);
			plinc(&dest,Startup->Throw[i]);	/* add the throw */
			draw_throw(&src,&dest,Startup->Throw[i]);

				/* Add 1 to src; if page boundary crossed
				   then handle it. */
			if (plinc(&src,1) != 0)
			{
				if (src.page >= nPages)
					break;
				start_page(src.page);
			}

			throwNum++;
		}
	}

		/* Cycle through the patterns */
	while ((Pattern != NULL) && (src.page < nPages) && (nCycles != 0))
	{
		for (P = Pattern; P != NULL; P = P->next)
		{
			if (Pattern->next != NULL)
				cycle_sym("newpatternsym", &src);
			else
				cycle_sym("newcyclesym", &src);

			for (i = 0; i < P->nReps; i++)
			{
				if (i != 0)
					cycle_sym("newcyclesym", &src);

				for (j = 0; j < P->nThrows; j++)
				{
					edit_check(&src, P->Throw[j]);

					plcopy(&dest,&src);
					plinc(&dest,P->Throw[j]);
					if (src.page < nPages)
						draw_throw(&src,&dest,P->Throw[j]);

						/* Add 1 to src; if page
						   boundary crossed then
						   handle it. */
					if (plinc(&src,1) != 0)
					{
						start_page(src.page);
					}

					throwNum++;
				}
			}
		}

		--nCycles;
	}

		/* Show the final page */
	showpage();
}

/*-------------------------------------------------------*/
void	read_comline(ac, av)

	int	ac;
	char	*av[];
{
	int	arg = 1;
	FILE	*inp;
	PATTERN	*P;

	while (arg < ac)
	{
		if ((mystricmp(av[arg],"-out") == 0) || ISALPHA(av[arg][0]))
		{
			if (av[arg][0] == '-')	/* skip switch */
				arg++;
			outfile = av[arg];
		}
		else if (mystricmp(av[arg],"-in") == 0)
		{
				/* Read patterns from input file */
			if (infile == NULL)
				infile = av[arg+1];
			read_file(av[arg+1]);

			arg++;
		}
		else if (mystricmp(av[arg],"-nops") == 0)
		{
			PSOut = 0;
		}

			/* Read a pattern if the switch is either
			   exactly "-p" or "-p" followed by a digit
			   or just a digit */
		else if (
		   (mystrnicmp(av[arg],"-p",2) == 0 && strlen(av[arg]) == 2)
		|| (mystrnicmp(av[arg],"-p",2) == 0 && ISDIGIT(av[arg][2]))
		|| ISDIGIT(av[arg][0]))
		{
			if (Pattern && (Pattern->nThrows == 0))
				P = Pattern;
			else
				P = new_pattern();

			if (av[arg][0] == '-')
			{
				if (ISDIGIT(av[arg][2]))
					P->nReps = atoi(&av[arg][2]);
				arg++;	/* skip switch */
			}

			for (; (arg < ac) && ISDIGIT(av[arg][0]); arg++)
			{
				P->Throw[P->nThrows] = atoi(av[arg]);
				P->nThrows++;
			}

			if (P != Pattern)
				add_pattern(P);

			--arg;	/* ... gets incremented at bottom of loop */
		}
		else if (mystricmp(av[arg],"-s") == 0)
		{
			for (arg++; (arg < ac) && ISDIGIT(av[arg][0]); arg++)
			{
				if (Startup == NULL)
					Startup = new_pattern();
				Startup->Throw[Startup->nThrows] = atoi(av[arg]);
				Startup->nThrows++;
			}

			--arg;	/* ... gets incremented at bottom of loop */
		}
		else if (mystricmp(av[arg],"-right") == 0)
		{
			init[R] = atoi(av[arg+1]);
			arg++;
		}
		else if (mystricmp(av[arg],"-left") == 0)
		{
			init[L] = atoi(av[arg+1]);
			arg++;
		}
		else if (mystricmp(av[arg],"-cycles") == 0)
		{
			nCycles = atoi(av[arg+1]);
			arg++;
		}
		else if (mystricmp(av[arg],"-nonums") == 0)
		{
			numLabs = 0;
		}
		else if (mystricmp(av[arg],"-noballnums") == 0)
		{
			ballNums = 0;
		}
		else if (mystricmp(av[arg],"-nomarks") == 0)
		{
			cycleMarks = 0;
		}
		else if (mystricmp(av[arg],"-noedit") == 0)
		{
			editMarks = 0;
		}
		else if (mystricmp(av[arg],"-nodebug") == 0)
		{
			debug = 0;
		}
		else if (mystricmp(av[arg],"-title") == 0)
		{
			if (title == NULL)
				title = av[arg+1];
			arg++;
		}
		else if (mystricmp(av[arg],"-cols") == 0)
		{
			sscanf(av[arg+1],"%d",&nCols);
			arg++;
		}
		else if (mystricmp(av[arg],"-rows") == 0)
		{
			sscanf(av[arg+1],"%d",&nRows);
			arg++;
		}
		else if (mystricmp(av[arg],"-pagesize") == 0)
		{
			sscanf(av[arg+1],"%f",&PageWidth);
			sscanf(av[arg+2],"%f",&PageHeight);
			arg += 2;
		}
		else if (mystricmp(av[arg],"-npages") == 0)
		{
			sscanf(av[arg+1],"%d",&nPages);
			arg++;
		}
		else if (mystricmp(av[arg],"-colwidth") == 0)
		{
			sscanf(av[arg+1],"%f",&ColWidth);
			arg++;
		}
		else if (mystricmp(av[arg],"-rowheight") == 0)
		{
			sscanf(av[arg+1],"%f",&RowHeight);
			arg++;
		}
		else if (mystricmp(av[arg],"-ballsize") == 0)
		{
			sscanf(av[arg+1],"%f",&BallSize);
			BallSize /= 100.0;	/* percentage */
			BallSize /= 2.0;	/* diameter to radius */
			arg++;
		}
		else if (mystricmp(av[arg],"-colour") == 0 ||
			 mystricmp(av[arg],"-color") == 0)
		{
			colour = 1;
		}
		else if (mystricmp(av[arg],"-shade") == 0)
		{
			shading = 1;
		}
		else if (mystricmp(av[arg],"-autostart") == 0)
		{
			autoStart = 1;
		}
		else if (mystricmp(av[arg],"-psfile") == 0)
		{
			psfile = (char *) malloc(strlen(av[arg+1])+1);
			strcpy(psfile, av[arg+1]);
			arg++;
		}
		else if (mystricmp(av[arg],"-pspath") == 0)
		{
			int len = strlen(av[arg+1]);

			pspath = (char *) malloc(len+2);
			strcpy(pspath, av[arg+1]);
			if (pspath[len-1] != DIR_SLASH)
			{
				pspath[len] = DIR_SLASH;
				pspath[len+1] = 0;
			}
			arg++;
		}
		else
		{
			fprintf(stderr,"unknown option %s\n", av[arg]);
			exit(1);
		}

		arg++;
	}
}

/*-------------------------------------------------------*/
void	read_file(fn)

	char	*fn;
{
	FILE	*inp;
	PATTERN	*P, *Q;
	char	*bp;
	int	i;
	int	ac;
	char	*av[BUF_SIZE/2];	/* bigger than really necessary */

	inp = fopen(fn, "r");
	if (inp == NULL)
	{
		fprintf(stderr, "couldn't open %s\n", fn);
		exit(1);
	}

		/* Files are read by converting each line into
		   a command line and then parsing the same way as
		   for the "real" command line. The only exception
		   is that lines starting with a '#' are treated
		   as comments. */
	while (!feof(inp))
	{
		if (fgets(buf, BUF_SIZE, inp) == NULL)
			break;

		ac = 1;		/* like regular cmd line, ignore first arg */
		if (buf[0] != '#')
		{
				/* Break up into words */
			bp = buf;
			while (*bp)	/* 0 at end of line */
			{
					/* Skip space */
				while (ISSPACE(*bp))
					bp++;

				if (*bp)
				{
					av[ac] = bp;
					ac++;
				}

					/* Skip word */
				while (*bp && !ISSPACE(*bp))
					bp++;

				if (*bp)
				{
						/* Mark end of word */
					*bp = 0;
					bp++;
				}
			}

			if (ac > 1)
				read_comline(ac, av);
		}
	}

	fclose(inp);
}

/*-------------------------------------------------------*/
PATTERN	*start_pattern(P)

	PATTERN	*P;
{
	PATTERN	*S = NULL;
	int	*landing;	/* for seeing where startup throws needed */
	int	nl;
	int	tmax = 0;	/* maximum throw */
	int	nb, nt;		/* number of balls, number of throws */
	int	err = 0;	/* error encountered */
	int	i, k;

	nt = P->nThrows;

	for (i = 0; i < nt; i++)
	{
		P->sumThrows += P->Throw[i];
		if (P->Throw[i] > tmax)
			tmax = P->Throw[i];
	}

	if ((nt > 0) &&  P->sumThrows%nt != 0)
	{
		if (debug)
			warning("throw total not divisible by number of throws");
		return(NULL);
	}
	else
	{
		P->nBalls = P->sumThrows/nt;
		nb = P->nBalls;
	}

		/* Check for more specific errors ... */

		/* Collisions
		 *	Th[i] - Th[(i+k)%nt] = k
		 * Nothrows
		 *	Th[i] > 0 && Th[(i+Th[i])%nt] = 0
		 * Ghostthrows
		 *	Th[i] > 0 && !Exist(k):Th[(i-k)%nt] = k
		 */
	for (i = 0; i < nt; i++)
	{
		if (P->Throw[i] > 0 && P->Throw[(i+P->Throw[i])%nt] == 0)
		{
			if (debug)
			{
				sprintf(buf, "throw %-d lands on a no-throw", i);
				warning(buf);
			}
			err = 1;
		}

		if (P->Throw[i] > 0)
		{
			for (k = 1; k <= tmax; k++)
				if (P->Throw[(i-k)%nt < 0 ? (i-k)%nt+nt : (i-k)%nt] == k)
					break;

			if (k > tmax)
			{
				if (debug)
				{
					sprintf(buf, "throw %-d has nothing to throw", i);
					warning(buf);
				}
				err = 1;
			}
		}

		for (k = 1; k <= tmax; k++)
		{
			if (P->Throw[i] - P->Throw[(i+k)%nt] == k)
			{
				if (i < (i+k)%nt)
				{
					if (debug)
					{
						sprintf(buf, "throws %-d and %-d land at the same time", i, (i+k)%nt);
						warning(buf);
					}
				}
				err = 1;
			}
		}
	}

	if (err)
		return(NULL);

		/* If the pattern has passed all these tests, then
		   it is self-sustaining, and we can generate a
		   startup pattern that will get it going. The simplest
		   way to do this is provide a pattern of length n
		   that puts the n balls into the right hand at the
		   right time (and the left hand too, of course!) */
	S = new_pattern();
	S->nBalls = nb;

		/* I don't know how big this array needs to be.
		   I'm presuming that at least one new ball is
		   needed each cycle. */
	nl = nb*nt + 1;
	landing = (int *)malloc(nl*sizeof(int));
	for(i = 0; i < nl; i++)
		landing[i] = 0;

	for (i = 0; i < nl && S->nThrows < nb; i++)
	{
		if (P->Throw[i%nt] > 0 && landing[i] == 0)
		{
			S->Throw[S->nThrows] = i - S->nThrows + nb;
			S->nThrows++;
		}

		if (i+P->Throw[i%nt] < nl)
			landing[i+P->Throw[i%nt]] = 1;
	}

	free(landing);

		/*** Debugging ***/
	if (i == nl)
		fprintf(stderr, "landings array not big enough!!\n");

		/* If the last throw == nb, then since the
		   first is >= nb, and each one is >= previous,
		   they must all == nb ... i.e. no startup
		   is required. */
	if (S->Throw[S->nThrows-1] == nb)
	{
		S->nThrows = 0;	/* not S = NULL, since this isn't an error */
	}

	return(S);
}

/*-------------------------------------------------------*/
void	add_pattern(P)

	PATTERN	*P;
{
	PATTERN	*Q;

	if (Pattern == NULL)
		Pattern = P;
	else
	{
		for (Q = Pattern; Q->next != NULL; Q = Q->next);
		Q->next = P;
	}
}

/*-------------------------------------------------------*/
void	plcopy(dest,src)

	PAGELOC	*dest, *src;
{
	dest->hand = src->hand;
	dest->row = src->row;
	dest->col = src->col;
	dest->page = src->page;
}

/*-------------------------------------------------------*/
int	plequal(dest,src)

	PAGELOC	*dest, *src;
{
	return(	(dest->hand == src->hand) &&
		(dest->row == src->row) &&
		(dest->col == src->col) &&
		(dest->page == src->page));
}

/*-------------------------------------------------------*/
int	plinc(pl,n)

	PAGELOC	*pl;
	int	n;
{
	int pageinc = 0;

	while (n > 0)
	{
		pl->hand = 1 - pl->hand;
		if (++pl->row >= nRows)
		{
			pl->row = 0;
			if (++pl->col >= nCols)
			{
				pl->col = 0;
				++pl->page;
				++pageinc;
			}
		}

		--n;
	}

	return(pageinc);	/* How many page boundaries? */
}

/*-------------------------------------------------------*/
void	start_page(n)

	int	n;
{
	BACKLOG	*b, *nxt;

		/* Don't do anything if all pages are done */
	if (n >= nPages)
		return;

	showpage();	/* Tell PostScript we're starting a new page */

	draw_page(n);	/* Draw in the grids, etc. */

		/* Do all the drawing from previous pages that overflowed */
	b = overflow[n];
	while (b != NULL)
	{
			/* Throw number is irrelevant, since start of
			   throw won't be drawn anyway. */
		draw_throw(&(b->src),&(b->dest),0);
		nxt = b->next;
		free(b);
		b = nxt;
	}

	overflow[n] = NULL;
}

/*-------------------------------------------------------*/
void	cycle_sym(name, pl)

	char	*name;
	PAGELOC	*pl;
{
	float	sx, sy, sr;

	if (throwNum >= nThrowsTotal || !cycleMarks)
		return;

	sr = RowHeight * 0.35;
	sx = ColRight[pl->col] - 2*sr;
	sy = ColTop - RowHeight*(pl->row+0.5);

	PSPRINT(printf("%-.2f %-.2f %-.2f %s\n", PSCOORD(sx),PSCOORD(sy),PSCOORD(sr),name));
}

/*-------------------------------------------------------*/
void	edit_sym(name, pl)

	char	*name;
	PAGELOC	*pl;
{
	float	sx, sy, sr;

	if (!editMarks)
		return;

	sr = RowHeight * 0.35;
	sx = ColHandX[pl->hand][pl->col];
	sy = ColTop - RowHeight*(pl->row+1+0.5);

	PSPRINT(printf("%-.2f %-.2f %-.2f %s\n", PSCOORD(sx),PSCOORD(sy),PSCOORD(sr),name));
}

/*-------------------------------------------------------*/
void	edit_check(pl,len)

	PAGELOC	*pl;
	int	len;
{
	if (throwNum >= nThrowsTotal)
		return;

	buf[0] = 0;

	if (BallPresent[throwNum] < 0)
	{
		edit_sym("collisionsym",pl);
		sprintf(buf,"collision at time %-d", throwNum+1);
	}
	else if (BallPresent[throwNum] == 0 && len != 0)
	{
		edit_sym("ghostballsym", pl);
		sprintf(buf,"nothing to throw at time %-d", throwNum+1);
	}
	else if (BallPresent[throwNum] > 0 && len == 0)
	{
		edit_sym("nothrowsym", pl);
		sprintf(buf,"left holding the bag at time %-d", throwNum+1);
	}

	if (debug && buf[0] != 0)
		warning(buf);

	if (throwNum+len < nThrowsTotal)
	{
		if (BallPresent[throwNum+len] == 0)
		{
			BallPresent[throwNum+len] = MAX(1,ABS(BallPresent[throwNum]));
		}
		else if (BallPresent[throwNum+len] > 0)
		{
				/* Collision! */
			BallPresent[throwNum+len] = -BallPresent[throwNum+len];
		}
	}
}

/*-------------------------------------------------------*/
PATTERN	*new_pattern()
{
	PATTERN	*p;

	p = (PATTERN *) malloc(sizeof(PATTERN));
	p->nBalls = 0;
	p->nThrows = 0;
	p->sumThrows = 0;
	p->nReps = 1;

	return(p);
}

/*-------------------------------------------------------*/
void	draw_page(n)

	int	n;	/* Page number (from 0) */
{
	int	i,j;
	float	bot;
	float	bw,bh;		/* width, height of boxes */
	float	bx;		/* x position of left box edge */
	float	yt,yb;		/* y position of top and bottom box */
	int	max_chars;

	linewidth(LINE_WIDTH);
	bot = ColTop-(float)nRows*RowHeight;

	bw = ColRight[nCols-1] - ColLeft[0];
	bh = TITLE_SPACE * 0.4;
	bx = ColLeft[0];
	yt = PageHeight - TOP_MARGIN - bh/2.0;
	yb = (ColTop + yt - bh/2.0)/2.0;

	max_chars = 75 * bw/PageWidth;

		/* Title */
	if (strlen(title) > max_chars - 5)
		title[max_chars - 5] = 0;

	box_text(bx,yt,bw,bh,title);

		/* Page number */
	sprintf(buf,"%-d", n+1);
	box_text(bx+bw-bh,yt,bh,bh,buf);

		/* And the pattern */
	i = 0;
	j = 0;
	buf[0] = 0;
	if (infile != NULL)
	{
		sprintf(buf,"%s: ", infile);
		i = strlen(buf);
	}

	if (Startup != NULL)
	{
		sprintf(&buf[i],"-s [ ");
		i = strlen(buf);

		while ((i < max_chars) && (j < Startup->nThrows))
		{
			sprintf(&buf[i],"%-d ",Startup->Throw[j]);
			i = strlen(buf);
			j++;
		}

		if (j == Startup->nThrows)
		{
			sprintf(&buf[i],"] ");
			i += 2;
		}
		else
		{
			sprintf(&buf[i],"...]");
			i += 4;
		}
	}

	j = 0;
	if (Pattern != NULL)
	{
		while ((i < max_chars) && (j < Pattern->nThrows))
		{
			sprintf(&buf[i],"%-d ",Pattern->Throw[j]);
			i = strlen(buf);
			j++;
		}

		if ((j < Pattern->nThrows) || (Pattern->next != NULL))
		{
			sprintf(&buf[i],"...");
			i += 3;
		}
	}
	
	box_text(bx,yb,bw,bh,buf);

		/* Draw the columns */
	for (i = 0; i < nCols; i++)
	{
		line(ColLeft[i],ColTop,ColLeft[i],bot);
		line(ColCenter[i],ColTop,ColCenter[i],bot);
		line(ColRight[i],ColTop,ColRight[i],bot);
	}

	for (i = 0; i <= nRows; i++)
	{
		for (j = 0; j < nCols; j++)
		{
			line(ColLeft[j],ColTop-i*RowHeight,ColRight[j],ColTop-i*RowHeight);
		}
	}
}

/*-------------------------------------------------------*/
void	draw_numbers(Pat)

	PATTERN	*Pat;
{
		/* Print the number of balls in first row of first column */
	if (Pat != NULL)
	{
		font("Times-Roman",RowHeight*0.65);
		if (init[R] >= 0)
		{
			sprintf(buf,"%-d", init[R]);
			text(ColHandX[R][0],ColTop-RowHeight/2.0,buf);
		}

		if (init[L] >= 0)
		{
			sprintf(buf,"%-d", init[L]);
			text(ColHandX[L][0],ColTop-RowHeight/2.0,buf);
		}
	}

		/* And draw circles around the numbers */
	circle(ColHandX[L][0],ColTop-RowHeight/2.0,RowHeight*0.4);
	circle(ColHandX[R][0],ColTop-RowHeight/2.0,RowHeight*0.4);
}

/*-------------------------------------------------------*/
void	draw_throw(src, dest, len)

	PAGELOC	*src, *dest;
	int	len;	/* redundant, but easier than recomputing */
{
	int	col, dcol, drow;
	float	sx,sy,dx,dy;
	float	xdrift;

		/* If page should be upside-down then compute
		   rows from bottom instead of top. */
	if (upsideDown)
	{
	}

		/* Label line, if requested */
	if (numLabs && (src->row >= 0) && (src->row <= nRows))
	{
		sprintf(buf,"%-d",throwNum+1);
		box_text(ColLeft[src->col]+RowHeight*0.1,
			 ColTop - RowHeight*(src->row+0.5),
			 RowHeight,RowHeight*0.8,buf);
	}

	if (plequal(src,dest))
		return;		/* Do nothing (i.e. a 0 throw) */

		/* Check if throw extends across page boundary */
	if (src->page != dest->page)
	{
			/* find "equivalent" destination in
			   last column of this page */
		dcol = nCols - 1;
		drow = dest->row + nRows*(dest->col+1 + nCols*(dest->page - src->page - 1));
		add_overflow(src,dest);
	}
	else
	{
		dcol = dest->col;
		drow = dest->row;
	}

		/* Line may extend across multiple columns.
		   The entire throw is redrawn in each possible
		   column, with clipping set to the area of
		   the particular column. */
	for (col = src->col; col <= dcol; col++)
	{
		sx = ColHandX[src->hand][col];
		sy = ColTop - RowHeight*(src->row-nRows*(col-src->col)+0.5);
		dx = ColHandX[dest->hand][col];
		dy = ColTop - RowHeight*(drow-nRows*(col-dcol)+0.5);

		clip(ColLeft[col],ColTop,ColRight[col],ColBot);

			/* An even throw comes back to the same
			   hand -- use a curve so that it won't
			   overlap other even throws. The curve
			   "drifts" away from vertical. */
		if (throwNum < nThrowsTotal)
			ballcolour(BallPresent[throwNum]);
		if (src->hand == dest->hand)
		{
			xdrift = (sx - ColCenter[col])*0.5;

				/* The (float) typecast on (sy-dy)
				   should be totally redundant, but
				   apparently some versions of cc
				   are braindead. */
			curve(sx,sy, sx+xdrift,sy - 0.3*(float)(sy-dy),
				sx+xdrift,dy + 0.3*(float)(sy-dy), dx,dy);
		}
		else
		{
			line(sx,sy,dx,dy);
		}

			/* Finally, draw the ball */
		drawball(sx,sy,BallSize);

		if (ballNums)
		{
			sprintf(buf,"%-d",len);
			font("Times-Bold",BallSize*1.5);
			outlinetext(sx,sy,buf);
		}

		unclip();
	}
}

/*-------------------------------------------------------*/
void	add_overflow(src,dest)

	PAGELOC	*src, *dest;
{
	BACKLOG	*b;
	int	page;

	page = src->page + 1;

		/* No point doing anything if it won't be seen anyway */
	if (page >= nPages)
		return;

	b = (BACKLOG *) malloc(sizeof(BACKLOG));

	b->src.hand = src->hand;
	b->src.row = src->row - nRows*(nCols - src->col);
	b->src.col = 0;
	b->src.page = page;

	b->dest.hand = dest->hand;
	b->dest.row = dest->row;
	b->dest.col = dest->col;
	b->dest.page = dest->page;

	b->next = overflow[page];
	overflow[page] = b;
}

/*-------------------------------------------------------*/
void	line(x1,y1,x2,y2)

	float	x1,y1,x2,y2;
{
	PSPRINT(printf("newpath %-.2f %-.2f moveto %-.2f %-.2f lineto stroke\n", PSCOORD(x1),PSCOORD(y1),PSCOORD(x2),PSCOORD(y2)));
}

/*-------------------------------------------------------*/
void	linewidth(w)

	float	w;
{
	PSPRINT(printf("%-f setlinewidth\n", w));
}

/*-------------------------------------------------------*/
void	circle(x,y,r)

	float	x,y,r;
{
	PSPRINT(printf("newpath %-.2f %-.2f %-f 0 360 arc stroke\n", PSCOORD(x),PSCOORD(y),PSCOORD(r)));
}

/*-------------------------------------------------------*/
void	curve(x1,y1,x2,y2,x3,y3,x4,y4)

	float	x1,y1,x2,y2,x3,y3,x4,y4;
{
	PSPRINT(printf("newpath %-.2f %-.2f moveto %-.2f %-.2f %-.2f %-.2f %-.2f %-.2f curveto stroke\n", PSCOORD(x1),PSCOORD(y1),PSCOORD(x2),PSCOORD(y2),PSCOORD(x3),PSCOORD(y3),PSCOORD(x4),PSCOORD(y4)));
}

/*-------------------------------------------------------*/
void	drawball(x,y,r)

	float	x,y,r;
{
	PSPRINT(printf("newpath %-.2f %-.2f %-f 0 360 arc fill 0.0 setgray\n",PSCOORD(x),PSCOORD(y),PSCOORD(r)));
}

/*-------------------------------------------------------*/
void	ballcolour(bnum)

	int	bnum;
{
	bnum = MAX(1,ABS(bnum));

	if (colour)
	{
		PSPRINT(printf("%-f 1.0 1.0 sethsbcolor\n",ballColour[bnum-1]));
	}
	else if (shading)
	{
		PSPRINT(printf("%-f setgray\n",ballColour[bnum-1]));
	}
	else
	{
		PSPRINT(printf("0.0 setgray\n"));
	}
}

/*-------------------------------------------------------*/
void	font(fn,size)

	char	*fn;
	float	size;	/* Size in inches */
{
	PSPRINT(printf("%-.2f /%s usefont\n", PSCOORD(size), fn));
}

/*-------------------------------------------------------*/
void	text(x,y,string)

	float	x,y;
	char	*string;
{
	PSPRINT(printf("%-.2f %-.2f moveto\n(%s) centertext\n", PSCOORD(x), PSCOORD(y), string));
}

/*-------------------------------------------------------*/
void	outlinetext(x,y,string)

	float	x,y;
	char	*string;
{
	PSPRINT(printf("%-.2f %-.2f moveto\n(%s) centeroutlinetext\n", PSCOORD(x), PSCOORD(y), string));
}

/*-------------------------------------------------------*/
void	box_text(x,y,width,height,string)

	float	x,y;
	float	width,height;
	char	*string;
{
	PSPRINT(printf("newpath %-.2f %-.2f moveto\n(%s) %-.2f %-.2f boxtext stroke\n", PSCOORD(x), PSCOORD(y), string, PSCOORD(width), PSCOORD(height)));
}

/*-------------------------------------------------------*/
void	clip(x1,y1,x2,y2)

	float	x1,y1,x2,y2;
{
	PSPRINT(printf("newpath %-.2f %-.2f moveto %-.2f %-.2f lineto %-.2f %-.2f lineto %-.2f %-.2f lineto closepath clip\n", PSCOORD(x1), PSCOORD(y1), PSCOORD(x2), PSCOORD(y1), PSCOORD(x2), PSCOORD(y2), PSCOORD(x1), PSCOORD(y2)));
}

/*-------------------------------------------------------*/
void	unclip()
{
	PSPRINT(printf("initclip\n"));
}

/*-------------------------------------------------------*/
void	showpage()
{
	PSPRINT(printf("showpage\n"));
}

/*-------------------------------------------------------*/
void	warning(string)

	char	*string;
{
	fprintf(stderr, "Warning: %s\n", string);
}

/*-------------------------------------------------------*/
#define	TOUPPER(C)	(('a' <= (C) && (C) <= 'z') ? ((C)-'a'+'A') : (C))

int	mystricmp(s1, s2)

	char	*s1, *s2;
{
	for ( ; TOUPPER(*s1) == TOUPPER(*s2); s1++, s2++)
	{
		if (*s1 == '\0')
			return(0);
	}

	return(*s1 - *s2);
}

/*-------------------------------------------------------*/
int	mystrnicmp(s1, s2, n)

	char	*s1, *s2;
	int	n;
{
	for (; n > 0 && TOUPPER(*s1) == TOUPPER(*s2); s1++, s2++, n--)
	{
		if (*s1 == '\0')
			return(0);
	}

	if (n == 0)
		return(0);
	else
		return(*s1 - *s2);
}
